from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="parabot",
    version="1.0.2",
    description="Execute robotframework test files/tests in paralel, even without special preparation of them",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/bednaJedna/parabot",
    author="bednaJedna",
    author_email="bednarik.radek@gmail.com",
    packages=find_packages(),
    install_requires=["robotframework", "selenium", "robotframework-seleniumLibrary"],
    license="MIT",
    keywords="robotframework, parallel execution, test, testing",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
)



import argparse
from datetime import datetime as dt
from pathlib import Path, PurePath
from typing import Any, List


def parse_args() -> Any:
    """Parsers arguments provided by user via CLI.
    
    Returns:
        Any -- object with parsed arguments as attributes
    
    See:
        https://docs.python.org/3/library/argparse.html
    """
    parser: Any = argparse.ArgumentParser(
        description="What tests do you want to run in parallel?"
    )
    parser.add_argument(
        "-a",
        "--all",
        action="store_true",
        help="Parallelize execution of all .robot files",
    )
    parser.add_argument(
        "-f",
        "--folders",
        action="extend",
        nargs="+",
        type=str,
        help="Enter relative path to specific folder with .robot files.",
    )
    parser.add_argument(
        "-t",
        "--tags",
        action="extend",
        nargs="+",
        type=str,
        help="Enter tags of tests/suites you want to run in parallel.",
    )
    parser.add_argument(
        "-to",
        "--timeout",
        action="store",
        type=int,
        help="Enter timeout in seconds if default 60 secs timeout for -f and -a args is too short for you.",
    )
    return parser.parse_args()


def get_parent_dir(test_filepath: Any) -> Any:
    """Returns PurePath to parent directory of argument filepath.
    
    Arguments:
        test_filepath {Any} -- python Path to .robot file
    
    Returns:
        Any -- python PurePath to parent directory.
    
    See:
        https://docs.python.org/3/library/pathlib.html
    """
    return PurePath(test_filepath).parent


def create_output_folder(basepath: Any, filename: str) -> Any:
    """Returns PurePath to new folder for report of <filename> provided as argument.

    Actual creation of the directory is handled by RobotFramework.
    
    Arguments:
        basepath {Any} -- PurePath to .robot file parent directory
        filename {str} -- .robot file filename
    
    Returns:
        Any -- PurePath to report folder of <filename> .robot file
    """
    timestamp: str = dt.now().strftime("%Y%m%d_%H%M%S")
    dirname: str = "_".join(["report", filename.rstrip(".robot"), timestamp])
    return PurePath(basepath, dirname)


def get_all_robot_files() -> List[Any]:
    """Returns list of Paths of all .robot files in the directory tree.
    
    Returns:
        List[Any] -- Paths of all .robot files in the directory tree.
    """
    return list(Path(".").rglob("*.robot"))


def get_specific_robot_files_by_paths(reldirpaths: List[str]) -> List[Any]:
    """Returns list of Paths of all .robot files in provided list of relative paths.
    
    Arguments:
        reldirpaths {List[str]} -- list of relative paths to .robot files
    
    Returns:
        List[Any] -- Paths to .robot files
    """
    output: List[Any] = []
    for path in reldirpaths:
        [output.append(path_) for path_ in list(Path(path).rglob("*.robot"))]
    return output



from io import StringIO
from multiprocessing import Process, get_context
from multiprocessing.context import TimeoutError
from typing import Any, Callable, List

from robot.run import run

from parabot.utils import (
    create_output_folder,
    get_all_robot_files,
    get_parent_dir,
    get_specific_robot_files_by_paths,
    parse_args,
)


def path_worker(filepath: Any) -> None:
    """Worker used by Pool.

    Runs instance of RobotFramework for given .robot file.
    
    Arguments:
        filepath {Any} -- python PurePath to .robot file to be executed by robot.run.run
    
    See:
        https://docs.python.org/3/library/pathlib.html
        https://robot-framework.readthedocs.io/en/v3.1.2/autodoc/robot.html#module-robot.run
    """
    basepath: Any = get_parent_dir(filepath)
    stdout: Any = StringIO()
    run(
        filepath, outputdir=create_output_folder(basepath, filepath.name), stdout=stdout
    )
    print(stdout.getvalue())
    stdout.close()


def tag_worker(tag: str) -> None:
    """Worker used by Process.

    Runs instance of RobotFramework for given tag.
    
    Arguments:
        tag {str} -- tag
    
    See:
        https://robot-framework.readthedocs.io/en/v3.1.2/autodoc/robot.html#module-robot.run
    """
    stdout: Any = StringIO()
    run(
        "./",
        outputdir=create_output_folder("./reports/", tag),
        include=[tag],
        stdout=stdout,
    )
    print(stdout.getvalue())
    stdout.close()


def pool_path_workers(
    path_worker: Callable, filepathslist: List[Any], timeout=60
) -> None:
    """Runs path_worker against list of arguments in parallel.

    Max. number of parallel processes is limited by no. of CPUs cores.
    
    Arguments:
        path_worker {Callable} -- path_worker function
        filepathslist {List[Any]} -- list of python PurePaths to .robot files
        timeout {int} -- timeout for async pool.map_async method. Can be altered by passing
        [-to <int>] or [--timeout <int>] CLI argument 
    """
    with get_context("spawn").Pool(maxtasksperchild=1) as p:
        try:
            p.map_async(path_worker, filepathslist).get(timeout)
        except TimeoutError as e:
            print(
                f"Your tests are running too long. Consider increasing the timeout via CLI parameter -to or --timeout.\n {str(e)}"
            )


def pool_tag_workers(tag_worker: Callable, tags: List[str]) -> None:
    """Runs tag_worker against list of tags in parallel.

    For each tag is spawned one process.
    
    Arguments:
        tag_worker {Callable} -- tag_worker function
        tags {List[str]} -- list of tags 
    """
    procesess: List[Any] = []

    for tag in tags:
        p: Any = Process(target=tag_worker, args=(tag,))
        p.start()
        procesess.append(p)
    # I am NOT really shure about this, but it works.
    # Will welcome any clarification or corrections.
    [proc.join() for proc in procesess]


def main() -> None:
    """Main function.

    See:
        https://stackoverflow.com/questions/30487767/check-if-argparse-optional-argument-is-set-or-not
    """
    filepathslist: List[Any]
    args: Any = parse_args()

    if args.all and (args.timeout is None):
        filepathslist = get_all_robot_files()
        pool_path_workers(path_worker, filepathslist)
    elif args.all and (args.timeout is not None):
        filepathslist = get_all_robot_files()
        pool_path_workers(path_worker, filepathslist, timeout=args.timeout)

    if (args.folders is not None) and (args.timeout is None):
        filepathslist = get_specific_robot_files_by_paths(args.folders)
        pool_path_workers(path_worker, filepathslist)
    elif (args.folders is not None) and (args.timeout is not None):
        filepathslist = get_specific_robot_files_by_paths(args.folders)
        pool_path_workers(path_worker, filepathslist, timeout=args.timeout)

    if args.tags is not None:
        pool_tag_workers(tag_worker, args.tags)


if __name__ == "__main__":
    main()
